import unittest
from itertools import zip_longest
from unittest.mock import patch

from terminaltables import SingleTable

from mentabotix.modules.botix import MovingTransition, MovingState


# Define a mock MovingState class for testing purposes


class TestMovingTransition(unittest.TestCase):

    def setUp(self):
        self.default_duration = 1.5

        def breker() -> bool:
            return True

        self.default_breaker = breker
        self.default_check_interval = 0.2
        self.default_from_state = MovingState(0)
        self.default_to_state = MovingState(0)

    def test_init_valid_input(self):
        # Test valid input with all parameters provided
        transition = MovingTransition(
            self.default_duration,
            self.default_breaker,
            self.default_check_interval,
            self.default_from_state,
            self.default_to_state,
        )
        self.assertEqual(transition.duration, self.default_duration)
        self.assertEqual(transition.breaker, self.default_breaker)
        self.assertEqual(transition.check_interval, self.default_check_interval)
        self.assertIsInstance(transition.from_states, list)
        self.assertIsInstance(transition.to_states, dict)

    def test_init_duration_not_positive(self):
        with self.assertRaises(ValueError):
            MovingTransition(-1, None, None, None, None)

    def test_init_from_states_valid_types(self):
        # Test valid from_states types
        transition = MovingTransition(self.default_duration, None, None, self.default_from_state, None)
        self.assertEqual(len(transition.from_states), 1)
        self.assertEqual(transition.from_states[0], self.default_from_state)

        from_states_iterable = [MovingState(0), MovingState(0)]
        transition = MovingTransition(self.default_duration, None, None, from_states_iterable, None)
        self.assertEqual(len(transition.from_states), len(from_states_iterable))
        self.assertEqual(transition.from_states, list(from_states_iterable))

    def test_init_from_states_invalid_type(self):
        with self.assertRaises(ValueError):
            MovingTransition(self.default_duration, None, None, "invalid", None)

    def test_init_to_states_valid_types(self):
        # Test valid to_states types
        transition = MovingTransition(self.default_duration, None, None, None, self.default_to_state)
        self.assertEqual(len(transition.to_states), 1)
        self.assertEqual(list(transition.to_states.values())[0], self.default_to_state)

        to_states_dict = {1: MovingState(0), 2: MovingState(1)}
        transition = MovingTransition(self.default_duration, None, None, None, to_states_dict)
        self.assertEqual(len(transition.to_states), len(to_states_dict))
        self.assertEqual(transition.to_states, to_states_dict)

    def test_init_to_states_invalid_type(self):
        with self.assertRaises(ValueError):
            MovingTransition(self.default_duration, None, None, None, "invalid")

    def test_add_from_state(self):
        transition = MovingTransition(self.default_duration, None, None, None, None)
        transition.add_from_state(self.default_from_state)
        self.assertEqual(len(transition.from_states), 1)
        self.assertEqual(transition.from_states[0], self.default_from_state)

    def test_add_to_state(self):
        transition = MovingTransition(self.default_duration, None, None, None, None)
        key = 1
        transition.add_to_state(key, self.default_to_state)
        self.assertEqual(len(transition.to_states), 1)
        self.assertEqual(transition.to_states[key], self.default_to_state)

    @patch("mentabotix.MovingTransition.tokenize")
    def test_tokenize_called(self, mock_tokenize):
        transition = MovingTransition(self.default_duration, None, None, None, None)
        transition.tokenize()
        mock_tokenize.assert_called_once_with()

    @patch("mentabotix.MovingTransition.clone")
    def test_clone_called(self, mock_clone):
        transition = MovingTransition(self.default_duration, None, None, None, None)
        transition.clone()
        mock_clone.assert_called_once_with()

    def test_clone_behavior(self):
        transition1 = MovingTransition(self.default_duration)
        transition2 = transition1.clone()
        self.assertEqual(transition1, transition2)
        self.assertNotEqual(id(transition1), id(transition2))
        self.assertNotEqual(id(transition1.to_states), id(transition2.to_states))
        self.assertNotEqual(id(transition1.from_states), id(transition2.from_states))

    def test_identifier(self):
        transition1 = MovingTransition(self.default_duration, None, None, None, None)
        transition2 = MovingTransition(self.default_duration, None, None, None, None)
        self.assertNotEqual(transition1.identifier, transition2.identifier)

    def test_str(self):
        from_states = [MovingState(100), MovingState(0)]
        to_states = {1: MovingState(4100), 2: MovingState(2000), 3: MovingState(0)}
        transition = MovingTransition(self.default_duration, None, None, from_states, to_states)
        temp = [["From", "To"]]
        for from_state, to_state in zip_longest(transition.from_states, transition.to_states.values()):
            temp.append([str(from_state) if from_state else "", str(to_state) if to_state else ""])

        self.assertEqual(str(transition), SingleTable(temp).table)

    def test_hash(self):
        transition1 = MovingTransition(self.default_duration, None, None, None, None)
        transition2 = MovingTransition(self.default_duration, None, None, None, None)
        self.assertNotEqual(hash(transition1), hash(transition2))


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock

from bdmc.modules.controller import CloseLoopController, MotorInfo

from mentabotix import MovingState


class TestMovingState(unittest.TestCase):

    def setUp(self):
        MovingState.__state_id_counter__ = 0

    def test_init_speeds(self):
        # Test FullPattern
        state_full = MovingState(50)
        self.assertEqual(state_full.unwrap(), (50, 50, 50, 50))

        # Test LRPattern
        state_lr = MovingState(30, 60)
        self.assertEqual(state_lr.unwrap(), (30, 30, 60, 60))

        # Test IndividualPattern
        state_individual = MovingState(10, 20, 30, 40)
        self.assertEqual(state_individual.unwrap(), (10, 20, 30, 40))

    def test_init_speed_expressions(self):
        used_context_variables = ["var1", "var2"]

        # Test FullExpressionPattern
        state_full_expr = MovingState(speed_expressions="var1+var2", used_context_variables=used_context_variables)
        self.assertEqual(state_full_expr.speed_expressions, ("var1+var2", "var1+var2", "var1+var2", "var1+var2"))

        # Test LRExpressionPattern
        state_lr_expr = MovingState(speed_expressions=("var1", "var2"), used_context_variables=used_context_variables)
        self.assertEqual(state_lr_expr.speed_expressions, ("var1", "var1", "var2", "var2"))

        # Test IndividualExpressionPattern
        state_indiv_expr = MovingState(
            speed_expressions=("var1", "var2", "var1+var2", "2*var1-var2"),
            used_context_variables=used_context_variables,
        )
        self.assertEqual(state_indiv_expr.speed_expressions, ("var1", "var2", "var1+var2", "2*var1-var2"))

    def test_init_invalid_arguments(self):
        with self.assertRaises(ValueError):
            MovingState()

        with self.assertRaises(ValueError):
            MovingState(1, 2, 3, 5, speed_expressions=("expr1", "expr2"))

        with self.assertRaises(ValueError):
            MovingState(speed_expressions=("expr1", "expr2"), used_context_variables=["var1", "var2"])

    def test_class_methods(self):
        # Test halt
        state_halt = MovingState.halt()
        self.assertEqual(state_halt.unwrap(), (0, 0, 0, 0))

        # Test straight
        state_straight_fwd = MovingState.straight(80)
        self.assertEqual(state_straight_fwd.unwrap(), (80, 80, 80, 80))
        state_straight_bwd = MovingState.straight(-80)
        self.assertEqual(state_straight_bwd.unwrap(), (-80, -80, -80, -80))

        # Test differential
        state_diff_l = MovingState.differential("l", radius=20, outer_speed=70)
        self.assertEqual(state_diff_l.unwrap(), (11, 11, 70, 70))
        state_diff_r = MovingState.differential("r", radius=20, outer_speed=70)
        self.assertEqual(state_diff_r.unwrap(), (70, 70, 11, 11))

        # Test turn
        state_turn_l = MovingState.turn("l", 90)
        self.assertEqual(state_turn_l.unwrap(), (-90, -90, 90, 90))
        state_turn_r = MovingState.turn("r", 90)
        self.assertEqual(state_turn_r.unwrap(), (90, 90, -90, -90))

        # Test drift
        state_drift_fl = MovingState.drift("fl", 50)
        self.assertEqual(state_drift_fl.unwrap(), (0, 50, int(76.5), 50))
        state_drift_rl = MovingState.drift("rl", 50)
        self.assertEqual(state_drift_rl.unwrap(), (50, 0, 50, int(76.5)))
        state_drift_rr = MovingState.drift("rr", 50)
        self.assertEqual(state_drift_rr.unwrap(), (int(76.5), 50, 0, 50))
        state_drift_fr = MovingState.drift("fr", 50)
        self.assertEqual(state_drift_fr.unwrap(), (50, int(76.5), 50, 0))

    def test_apply(self):
        state = MovingState(20, 30, 40, 50)
        state_applied = state.apply(1.5)

        self.assertEqual(state_applied.unwrap(), (30, 45, 60, 75))

    def test_clone(self):
        original_state = MovingState(10, 20, 30, 40)
        cloned_state = original_state.clone()
        print()

        print(cloned_state)
        print(original_state)
        print()
        self.assertEqual(cloned_state.unwrap(), (10, 20, 30, 40))
        self.assertIsNot(cloned_state, original_state)  # Ensure a new instance is returned

        original_state = MovingState(speed_expressions="var1", used_context_variables=["var1"])

        cloned_state = original_state.clone()
        print(cloned_state)
        print(original_state)
        print()

        self.assertEqual(cloned_state.speed_expressions, ("var1", "var1", "var1", "var1"))
        self.assertIsNot(cloned_state, original_state)  # Ensure a new instance is returned()
        self.assertNotEqual(id(original_state.after_exiting), id(cloned_state.after_exiting))
        self.assertNotEqual(id(original_state.before_entering), id(cloned_state.before_entering))

    def test_tokenize(self):

        con_mock = Mock(
            spec=CloseLoopController(
                [MotorInfo(1), MotorInfo(2), MotorInfo(3), MotorInfo(4)], port="COM4", context={"var1": 10, "var2": 20}
            )
        )

        state_with_speeds = MovingState(10, 20, 30, 40)
        tokens, context = state_with_speeds.tokenize(con_mock)
        self.assertEqual(tokens, [".set_motors_speed((10, 20, 30, 40))"])
        self.assertEqual(context, {})  # No context needed for states with speeds

        state_with_expressions = MovingState(
            speed_expressions=("var1", "var2", "var1+var2", "2*var1-var2"),
            used_context_variables=["var1", "var2"],
        )
        tokens, context = state_with_expressions.tokenize(con_mock)
        self.assertCountEqual(
            [
                ".set_motors_speed(((state1_context_getter_temp_2:=state1_context_getter_2()), "
                "(state1_context_getter_temp_1:=state1_context_getter_1()), "
                "state1_context_getter_temp_2+state1_context_getter_temp_1, "
                "2*state1_context_getter_temp_2-state1_context_getter_temp_1))"
            ][0],
            tokens[0],
        )

        self.assertIn("state1_context_getter_1", context)
        self.assertIn("state1_context_getter_2", context)

        print(f"Tokens: {tokens}, Context: {context}")
        # Tokens and context assertions depend on the NameGenerator implementation, which is not provided
        # You can add appropriate assertions here once you have the actual generated tokens and context

    def test_hash_and_eq(self):
        state1 = MovingState(10, 20, 30, 40)
        print(f"State1: {state1}")
        state2 = MovingState(10, 20, 30, 40)
        print(f"State2: {state2}")
        state3 = MovingState(11, 20, 30, 40)
        print(f"State3: {state3}")
        self.assertEqual(hash(state1), hash(state2) - 1)
        self.assertNotEqual(hash(state1), hash(state3))

        self.assertEqual(state1, state2)
        self.assertNotEqual(state1, state3)

    def test_str(self):
        state = MovingState(10, 20, 30, 40)
        self.assertEqual(str(state), f"{state.state_id}-MovingState(10, 20, 30, 40)")
        state = MovingState(speed_expressions="var1", used_context_variables=["var1"])
        self.assertEqual(str(state), f"{state.state_id}-MovingState('var1', 'var1', 'var1', 'var1')")

    def test_random_turn(self):
        from mentabotix import MovingTransition, Botix

        con = CloseLoopController(
            [MotorInfo(1), MotorInfo(2), MotorInfo(3), MotorInfo(4)], port="COM10", context={"var1": 10, "var2": 20}
        ).start_msg_sending()
        state = MovingState.rand_turn(con, 500)
        print(state)
        end = MovingState(0)
        trans = MovingTransition(1, from_states=[state], to_states={0: end})

        b = Botix(controller=con, token_pool=[trans])

        fun = b.compile()
        fun()
        con.stop_msg_sending()


if __name__ == "__main__":
    unittest.main()



from .modules.botix import MovingState, MovingTransition, Botix
from .modules.exceptions import BadSignatureError, RequirementError, SamplerTypeError, TokenizeError, StructuralError
from .modules.logger import set_log_level
from .modules.menta import Menta, SequenceSampler, IndexedSampler, DirectSampler, SamplerUsage, SamplerType, Sampler

from .tools.composers import MovingChainComposer, straight_chain, snaking_chain, scanning_chain, random_lr_turn_branch
from .tools.generators import NameGenerator, Multipliers, make_multiplier_generator
from .vision.tagdetector import TagDetector

__all__ = [
    "set_log_level",
    # botix
    "MovingState",
    "MovingTransition",
    "Botix",
    # menta
    "Menta",
    "SequenceSampler",
    "IndexedSampler",
    "DirectSampler",
    "SamplerUsage",
    "SamplerType",
    "Sampler",
    # vision
    "TagDetector",
    # exceptions
    "BadSignatureError",
    "RequirementError",
    "SamplerTypeError",
    "TokenizeError",
    "StructuralError",
    # tools/composers
    "MovingChainComposer",
    "straight_chain",
    "snaking_chain",
    "scanning_chain",
    "random_lr_turn_branch",
    # tools/generators
    "NameGenerator",
    "Multipliers",
]



import random
import unittest

from bdmc import CloseLoopController, MotorInfo

from mentabotix import MovingChainComposer, MovingState, MovingTransition, straight_chain, Botix, random_lr_turn_branch
from mentabotix.tools.composers import __PLACE_HOLDER__


# Assuming MovingState, MovingTransition, UnitType, StateTransitionPack are defined elsewhere
# You'll need to replace these with the actual classes and types


class TestComposer(unittest.TestCase):
    def setUp(self):
        self.moving_chain_composer = MovingChainComposer()

    def test_last_state(self):
        # Test when there are no states
        self.assertIsNone(self.moving_chain_composer.last_state)

        # Test with one state
        state1 = MovingState(10)
        self.moving_chain_composer.add(state1)
        self.assertEqual(self.moving_chain_composer.last_state, state1)

        # Test with multiple states
        state2 = MovingState(100)
        with self.assertRaises(ValueError):
            self.moving_chain_composer.add(state2)

        tran1 = MovingTransition(1)
        tran2 = MovingTransition(2)
        self.moving_chain_composer.add(tran1)
        with self.assertRaises(ValueError):
            self.moving_chain_composer.add(tran2)
        # Test with one state
        state3 = MovingState(1052)
        self.moving_chain_composer.add(state3)
        self.assertEqual(self.moving_chain_composer.last_state, state3)

    def test_last_transition(self):
        # Test when there are no transitions
        self.assertIsNone(self.moving_chain_composer.last_transition)
        self.moving_chain_composer.add(MovingState(0))
        # Test with one transition
        tran1 = MovingTransition(10)
        self.moving_chain_composer.add(tran1)
        self.assertEqual(self.moving_chain_composer.last_transition, tran1)

        # Test with multiple transitions
        tran2 = MovingTransition(100)
        with self.assertRaises(ValueError):
            self.moving_chain_composer.add(tran2)

        state1 = MovingState(1)
        state2 = MovingState(2)
        self.moving_chain_composer.add(state1)
        with self.assertRaises(ValueError):
            self.moving_chain_composer.add(state2)
        # Test with one transition
        tran3 = MovingTransition(1052)
        self.moving_chain_composer.add(tran3)
        self.assertEqual(self.moving_chain_composer.last_transition, tran3)

    def test_next_need(self):
        # Test initial next need
        self.assertEqual(self.moving_chain_composer.next_need, MovingState)

        # Test after adding a MovingState
        self.moving_chain_composer.add(MovingState(0))
        self.assertEqual(self.moving_chain_composer.next_need, MovingTransition)

    def test_init_container(self):
        self.moving_chain_composer.add(MovingState(0))
        self.moving_chain_composer.init_container().add(MovingState(0))

    def test_export_structure(self):
        # Test with no units
        self.assertEqual(self.moving_chain_composer.export_structure(), ([], []))

        # Test with states and transitions
        state1, state2 = MovingState(1), MovingState(2)
        transition1, transition2 = MovingTransition(1), MovingTransition(2)
        (self.moving_chain_composer.add(state1).add(transition1).add(state2).add(transition2))

        expected_structure = ([state1, state2], [transition1, transition2])
        self.assertEqual(self.moving_chain_composer.export_structure(), expected_structure)

    def test_add(self):
        # Test adding correct unit types
        state = MovingState(1)
        transition = MovingTransition(1)
        self.moving_chain_composer.add(state)
        self.moving_chain_composer.add(transition)

        # Test adding incorrect unit type
        with self.assertRaises(ValueError):
            self.moving_chain_composer.add(14)

    # 测试没有breaker的情况
    def test_straight_chain_without_breaker(self):
        start_speed = 50
        end_speed = 100
        duration = 5.0
        power_exponent = 1.0
        interval = 0.07

        # 调用待测试函数
        states, transitions = straight_chain(
            start_speed=start_speed,
            end_speed=end_speed,
            duration=duration,
            power_exponent=power_exponent,
            interval=interval,
        )

        # 断言判断结果是否符合预期
        self.assertEqual(len(states), int(duration / interval) + 1)
        self.assertEqual(states[0].unwrap()[0], start_speed)
        self.assertEqual(states[-1].unwrap()[0], end_speed)
        for i in range(len(transitions) - 1):
            self.assertIsInstance(transitions[i], MovingTransition)
            self.assertEqual(transitions[i].duration, interval)

    # 测试有breaker函数的情况
    def test_straight_chain_with_breaker(self):
        start_speed = 50
        end_speed = 100
        duration = 5.0
        power_exponent = 1.0
        interval = 0.1

        def break_function() -> bool:
            return random.random() < 0.1

        # 调用待测试函数
        states, transitions = straight_chain(
            start_speed=start_speed,
            end_speed=end_speed,
            duration=duration,
            power_exponent=power_exponent,
            interval=interval,
            breaker=break_function,
        )

        # 断言判断结果是否符合预期
        self.assertTrue(any(isinstance(t, MovingTransition) and t.breaker for t in transitions))
        self.assertEqual(int(duration / interval), len(transitions))
        self.assertAlmostEqual(duration, sum(t.duration for t in transitions))
        self.assertEqual(int(duration / interval) + 1 + 1, len(states))  # 包含了break状态
        self.assertIn(MovingState(0), states)  # 确认state_on_break被加入到states列表中

    def test_exp(self):
        start_speed = 50
        end_speed = 100
        duration = 5.0
        power_exponent = 2.0
        interval = 0.1

        def break_function() -> bool:
            return random.random() < 0.1

        # 调用待测试函数
        states, transitions = straight_chain(
            start_speed=start_speed,
            end_speed=end_speed,
            duration=duration,
            power_exponent=power_exponent,
            interval=interval,
            breaker=break_function,
        )

    def test_structure(self):
        start_speed = 50
        end_speed = 100
        duration = 5.0
        power_exponent = 2.0
        interval = 0.1

        def break_function() -> bool:
            return random.random() < 0.1

        # 调用待测试函数
        states, transitions = straight_chain(
            start_speed=start_speed,
            end_speed=end_speed,
            duration=duration,
            power_exponent=power_exponent,
            interval=interval,
            breaker=break_function,
        )

        botix = Botix(controller=CloseLoopController([MotorInfo(i) for i in range(4)]), token_pool=transitions)
        botix.export_structure("long_chain.puml", botix.token_pool)

    # 测试breaker不是None且不是callable的情况
    def test_straight_chain_with_invalid_breaker(self):
        start_speed = 50
        end_speed = 100
        duration = 5.0
        power_exponent = 1.0
        interval = 0.07

        # 使用错误的breaker参数
        with self.assertRaises(ValueError):
            straight_chain(
                start_speed=start_speed,
                end_speed=end_speed,
                duration=duration,
                power_exponent=power_exponent,
                interval=interval,
                breaker="not a callable",
            )

    def test_mk(self):
        from mentabotix import MovingChainComposer, MovingState, MovingTransition

        # init the state-transition composer
        comp = MovingChainComposer()

        # add some states and transitions one by one to the composer, the composer will auto-connect the states and transitions
        (
            comp.add(MovingState(0))
            .add(MovingTransition(0.2))
            .add(MovingState(1000))
            .add(MovingTransition(0.3))
            .add(MovingState(2000))
        )

        # export the structure
        states, transitions = comp.export_structure()

        # let's use botix to make the visualization!
        # first make the botix object
        con = CloseLoopController(motor_infos=[MotorInfo(i) for i in range(4)])
        botix = Botix(controller=con, token_pool=transitions)

        # make the visualization
        botix.export_structure("composed.puml", botix.token_pool)

    def test_valid_input(self):

        # 定义测试用的变量
        start_state = MovingState(100)
        end_state = MovingState(500)
        start_state_duration = 2.0
        turn_speed = 5000
        turn_duration = 1.5
        turn_left_prob = 0.5

        # 正常输入的测试
        start_transition, turn_transition = random_lr_turn_branch(
            start_state, end_state, start_state_duration, turn_speed, turn_duration, turn_left_prob
        )

        Botix.export_structure("random_lr_turn_branch.puml", [start_transition, turn_transition])
        # 检查 start_transition 是否正确
        self.assertIsInstance(start_transition, MovingTransition)
        self.assertEqual(start_transition.from_states[0], start_state)
        self.assertIn(False, start_transition.to_states)
        self.assertIn(True, start_transition.to_states)

        # 检查 turn_transition 是否正确
        self.assertIsInstance(turn_transition, MovingTransition)
        for a, b in zip(
            turn_transition.from_states, [MovingState.turn("l", turn_speed), MovingState.turn("r", turn_speed)]
        ):
            self.assertEqual(a, b)
        self.assertEqual(turn_transition.to_states[__PLACE_HOLDER__], end_state)
        self.assertEqual(turn_transition.duration, turn_duration)

    def test_invalid_turn_speed(self):
        # 测试无效的 turn_speed 输入
        with self.assertRaises(ValueError):
            random_lr_turn_branch(MovingState(0), MovingState(20), 2.0, -1, 1.5, 1.5)  # Invalid turn_speed


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from inspect import signature, Signature
from typing import (
    Callable,
    TypeAlias,
    List,
    Optional,
    Tuple,
    Sequence,
    NamedTuple,
    Union,
    Self,
    SupportsInt,
    SupportsFloat,
    Dict,
    Any,
)

from .exceptions import BadSignatureError, RequirementError
from .logger import _logger

SensorData: TypeAlias = float | int
# basically, no restrictions, support py objects or ctypes._CData variants
SensorDataSequence: TypeAlias = Sequence[SensorData]
UpdaterClosure = Callable[[], SensorDataSequence] | Callable[[], SensorData]

SequenceSampler: TypeAlias = Callable[[], SensorDataSequence]
IndexedSampler: TypeAlias = Callable[[int], SensorData]
DirectSampler: TypeAlias = Callable[[], SensorData]

Sampler: TypeAlias = Union[SequenceSampler, IndexedSampler, DirectSampler]
SourceCode: TypeAlias = str


class SamplerType(Enum):
    """
    采样器类型
    """

    SEQ_SAMPLER: int = 1
    IDX_SAMPLER: int = 2
    DRC_SAMPLER: int = 3


class SamplerUsage(NamedTuple):
    used_sampler_index: int
    required_data_indexes: Sequence[int]


class Menta:
    # reserved to check the return type of the sampler
    __supported__ = (SupportsInt, SupportsFloat)

    def __init__(
        self,
        samplers: Optional[List[Sampler]] = None,
    ):
        self.samplers: List[Sampler] = samplers or []
        self.sampler_types: List[SamplerType] = []
        self.update_sampler_types()

    def update_sampler_types(self) -> Self:
        """
        更新采样器类型列表。遍历self.samplers中的每个采样器，根据其返回类型和参数数量将其分类为序列采样器（SEQ_SAMPLER）、
        索引采样器（IDX_SAMPLER）或直接响应采样器（DRC_SAMPLER），并更新self.sampler_types列表。
        如果采样器没有指定返回类型或指定的返回类型不被支持，则抛出异常。

        返回值:
            Self: 更新后的实例自身。
        """

        self.sampler_types.clear()  # 清空当前的采样器类型列表
        for sampler in self.samplers:  # 遍历所有采样器
            sig = signature(sampler)  # 获取采样器的签名
            try:
                match sig.return_annotation:  # 根据采样器的返回类型进行匹配
                    case seq_sampler_ret if hasattr(seq_sampler_ret, "__getitem__") and len(sig.parameters) == 0:
                        # 如果返回类型是序列且无参数，则认定为序列采样器
                        self.sampler_types.append(SamplerType.SEQ_SAMPLER)
                    case idx_sampler_ret if len(sig.parameters) == 1 and isinstance(
                        idx_sampler_ret, self.__supported__
                    ):
                        # 如果返回类型是索引且只有一个参数，则认定为索引采样器
                        self.sampler_types.append(SamplerType.IDX_SAMPLER)
                    case drc_sampler_ret if len(sig.parameters) == 0 and isinstance(
                        drc_sampler_ret, self.__supported__
                    ):
                        # 如果返回类型是直接响应且无参数，则认定为直接响应采样器
                        self.sampler_types.append(SamplerType.DRC_SAMPLER)
                    case Signature.empty:
                        # 如果采样器没有指定返回类型，则抛出异常
                        raise BadSignatureError(
                            f"Sampler {sampler} must have annotated return type!\nGot {sig.return_annotation}"
                        )
                    case invalid_sampler_ret:
                        # 如果采样器的返回类型注解无效，则抛出异常
                        raise BadSignatureError(
                            f"Sampler {sampler} has invalid return type annotation(s)!\n"
                            f"Must be {SensorDataSequence} or {SensorData} but got {invalid_sampler_ret}"
                        )
            except TypeError:
                raise BadSignatureError(f"Sampler {sampler} has invalid signature!\nGot {sig}")
        return self  # 返回更新后的实例自身

    def construct_updater(self, usages: List[SamplerUsage]) -> Callable[[], Tuple] | Callable[[], SensorData]:
        """
        Constructs an updater function based on the given list of sampler usages.

        Args:
            usages (List[SamplerUsage]): A list of sampler usages.

        Returns:
            Callable[[], Tuple]: The constructed updater function.

        Raises:
            ValueError: If the number of sampler usages does not match the number of samplers.
            RuntimeError: If an unsupported sampler type is encountered.
        """
        self.update_sampler_types()
        if len(self.sampler_types) != len(self.samplers):
            raise ValueError(
                f"Number of sampler usages ({len(usages)}) does not match number of samplers ({len(self.samplers)}), have you used the update_sampler_types() method?"
            )
        if len(usages) == 0:
            raise RequirementError("Can't resolve the empty Usage List")
        update_funcs: List[UpdaterClosure] = []

        for usage in usages:
            _logger.debug(
                f"Sampler_type: {self.sampler_types[usage.used_sampler_index]}|Required: {usage.required_data_indexes}"
            )
            sampler = self.samplers[usage.used_sampler_index]
            sampler_type = self.sampler_types[usage.used_sampler_index]
            match sampler_type:
                case SamplerType.SEQ_SAMPLER:
                    update_funcs.append(self.resolve_seq_sampler(sampler, usage.required_data_indexes))
                case SamplerType.IDX_SAMPLER:
                    update_funcs.append(self.resolve_idx_sampler(sampler, usage.required_data_indexes))
                case SamplerType.DRC_SAMPLER:
                    update_funcs.append(self.resolve_drc_sampler(sampler, usage.required_data_indexes))
                case _:
                    raise RuntimeError(f"Unsupported sampler type: {sampler_type}")

        # TODO allow make flatten ret
        if len(update_funcs) == 1:
            return update_funcs[0]
        eval_kwargs = {f"func_{i}": update_funcs[i] for i in range(len(update_funcs))}
        func_call_strings = [f"{func}()" for func in eval_kwargs]
        eval_string = "lambda:" + "(" + f",".join(func_call_strings) + ")"
        eval_obj = eval(eval_string, eval_kwargs)
        return eval_obj

    def construct_inlined_function(
        self,
        usages: List[SamplerUsage],
        judging_source: List[SourceCode] | SourceCode,
        extra_context: Dict[str, Any] = None,
        return_type_varname: Optional[str] = None,
        return_raw: bool = False,
    ) -> Callable[[], Any] | Tuple[str, Dict[str, Any]]:
        """
        构建一个内联函数，根据给定的采样器使用情况、判断源代码和额外上下文信息。

        Args:
            usages: 一个SamplerUsage对象列表，描述了如何使用采样器。
            judging_source: 判断源代码，可以是单个SourceCode对象或其列表。这些代码将被嵌入到生成的函数中。
            extra_context: 一个可选的字典，包含执行判断源代码时需要的额外上下文变量。
            return_type_varname: 一个可选的字符串，指定生成函数的返回类型变量名。如果提供，必须在extra_context中定义。
            return_raw: 一个布尔值，指示是否返回未执行的函数源码和使用的采样器信息。

        Returns:
            如果return_raw为True，则返回函数源码和使用的采样器信息的元组；
            否则，返回一个可调用对象，该对象在执行时根据判断源代码和提供的采样器数据返回结果。

        Raises:
            RequirementError: 如果返回类型变量名在extra_context中未定义，或者判断源中缺少必要的占位符。
            RuntimeError: 如果遇到了不支持的采样器类型。
        """
        if return_type_varname and return_type_varname not in extra_context:
            raise RequirementError(
                f'The return_type_varname "{return_type_varname}" is not defined in extra_context: {extra_context}.'
            )

        # 将judging_source统一处理为字符串格式
        judging_source: str = judging_source if isinstance(judging_source, str) else "\n ".join(judging_source)

        # 定义返回标识符，检查判断源中是否包含该标识符
        RET_IDENTIFIER: str = "ret"
        if RET_IDENTIFIER not in judging_source:
            raise RequirementError(
                f"Can't find {RET_IDENTIFIER} in judging source, you must define a variable named {RET_IDENTIFIER} in your judging source!\n"
                f"Since it will be used as the return value of the judge function."
            )

        # 匹配使用情况与相应的采样器和采样器类型
        _logger.debug("Match Usage with corresponding samplers and sampler_types")
        compile_context: Dict[str, Any] = {
            f"func_{i}": self.samplers[usage.used_sampler_index] for i, usage in enumerate(usages)
        }
        _logger.debug(f"Matched samplers: {compile_context}")
        self.update_sampler_types()
        used_sampler_types = [self.sampler_types[usage.used_sampler_index] for usage in usages]
        _logger.debug(f"Matched sampler_types: {used_sampler_types}")

        # 根据采样器类型，为采样器数据创建索引表达式
        sampler_temp_var_names_mapping: Dict[str, str] = {}
        indexed_expressions: List[str] = []
        for usage, sampler_type, func_name in zip(usages, used_sampler_types, compile_context):

            match sampler_type:
                case SamplerType.SEQ_SAMPLER:
                    sampler_temp_var_names_mapping[temp_name := f"{func_name}_temp"] = func_name
                    indexed_expressions.extend(self._index_for_seq_sampler_data(temp_name, usage.required_data_indexes))
                case SamplerType.IDX_SAMPLER:
                    indexed_expressions.extend(self._index_for_idx_sampler_data(func_name, usage.required_data_indexes))
                case SamplerType.DRC_SAMPLER:
                    sampler_temp_var_names_mapping[temp_name := f"{func_name}_temp"] = func_name
                    indexed_expressions.extend(self._index_for_drc_sampler_data(temp_name, usage.required_data_indexes))
                case _:
                    raise RuntimeError(f"Unsupported sampler type: {sampler_type}")
        _logger.debug(f"Created {len(indexed_expressions)} indexed expressions.")

        # 创建占位变量并检查是否所有占位符都被包含在判断源中
        placebo_var_names = [f"s{i}" for i in range(len(indexed_expressions))]
        _logger.debug(f"Created {len(placebo_var_names)} placebo variables.")
        _logger.debug("Checking that all placeholders are included in judging_source")
        if not_included := [placebo for placebo in placebo_var_names if placebo not in judging_source]:
            raise RequirementError(
                f"Judging source must have all placeholders: {placebo_var_names} in judging_source\n"
                f"Missing: {not_included}"
            )

        # 替换判断源中的占位符为索引表达式
        for placebo, expr in zip(placebo_var_names, indexed_expressions):
            _logger.debug(f'Replacing "{placebo}" with "{expr}".')
            judging_source = judging_source.replace(placebo, expr)

        # 创建临时变量的源代码，用于在执行环境中存储采样器的返回值
        temp_var_source: str = (
            ",".join(sampler_temp_var_names_mapping.keys())
            + "="
            + ",".join([f"{fname}()" for fname in sampler_temp_var_names_mapping.values()])
        )
        _logger.debug(f"Created temp_var_source: {temp_var_source}")

        # 构建完整的函数源码并编译执行
        function_head: str = f"def _func() -> {return_type_varname}:" if return_type_varname else "def _func():"
        func_source = f"{function_head}\n" f" {temp_var_source}\n" f" {judging_source}\n" f" return {RET_IDENTIFIER}"
        _logger.debug(f"Created func_source: {func_source}")
        _logger.debug("Compiling func_source")

        # 更新执行环境中的采样器和额外上下文信息

        compile_context.update(extra_context) if extra_context else None
        if return_raw:
            return func_source, compile_context
        exec(func_source, compile_context)  # exec the source with the context
        func_obj: Callable[[], bool] = compile_context.get("_func")
        _logger.debug(f"Succeed, compiled func_obj: {func_obj}")
        return func_obj

    @staticmethod
    def _index_for_seq_sampler_data(data_var_name: str, required: List[int] | int) -> List[str]:
        """
        A function that generates a list of indexed expressions based on the given data variable name and a list of required indexes.

        Args:
            data_var_name (str): The name of the data variable.
            required (List[int] | int): Either a single integer representing the required sequence length or a list of required data indexes.

        Returns:
            List[str]: A list of indexed expressions based on the data variable name and required indexes.
        """
        match required:
            case int(required_seq_length):
                final_required_data_indexes = range(required_seq_length)
            case list(required_data_indexes):
                final_required_data_indexes = required_data_indexes
            case _:
                raise RequirementError(
                    f"Unknown Input, arg::required has to be either list[int] or int, got {required}"
                )
        return [f"{data_var_name}[{i}]" for i in final_required_data_indexes]

    @staticmethod
    def _index_for_drc_sampler_data(data_var_name: str, required: List[int] | int) -> List[str]:
        """
        A function that generates a list of indexed expressions based on the given data variable name and a list of required indexes.

        Args:
            data_var_name (str): The name of the data variable.
            required (List[int] | int): Either a single integer representing the required sequence length or a list of required data indexes.

        Returns:
            List[str]: A list of indexed expressions based on the data variable name and required indexes.
        """
        match required:
            case []:
                raise RequirementError("Can't resolve the empty Usage List")
            case int(required_seq_length):
                final_required_data_indexes = range(required_seq_length)
            case list(required_data_indexes):
                final_required_data_indexes = required_data_indexes
            case _:
                raise RequirementError(
                    f"Unknown Input, arg::required has to be either list[int] or int, got {required}"
                )
        return [f"(({data_var_name}>>{i})&1)" for i in final_required_data_indexes]

    @staticmethod
    def _index_for_idx_sampler_data(func_var_name: str, required: List[int]) -> List[str]:
        """
        Generate a list of indexed expressions based on the given function variable name and a list of required indexes.

        Args:
            func_var_name (str): The name of the function variable.
            required (List[int]): A list of required indexes.

        Returns:
            List[str]: A list of indexed expressions based on the function variable name and required indexes.
        """
        return [f"{func_var_name}({i})" for i in required]

    @staticmethod
    def resolve_seq_sampler(sampler: SequenceSampler, required_data_indexes: Sequence[int]) -> UpdaterClosure:
        """
        Resolves the sampler based on the given sequence sampler and required data indexes.

        Args:
            sampler (SequenceSampler): The sequence sampler to resolve.
            required_data_indexes (Sequence[int]): The required data indexes.

        Returns:
            UpdaterClosure: A callable that returns a tuple of SensorData objects
            or a single SensorData object based on the number of required data indexes.

        Raises:
            None

        Examples:
            sampler = SequenceSampler()
            required_data_indexes = []
            resolved_sampler = resolve_seq_sampler(sampler, required_data_indexes)
            data = resolved_sampler()  # Returns a tuple of all SensorData objects

            sampler = SequenceSampler()
            required_data_indexes = [0]
            resolved_sampler = resolve_seq_sampler(sampler, required_data_indexes)
            data = resolved_sampler()  # Returns the SensorData object at index 0

            sampler = SequenceSampler()
            required_data_indexes = [0, 1, 2]
            resolved_sampler = resolve_seq_sampler(sampler, required_data_indexes)
            data = resolved_sampler()  # Returns a tuple of SensorData objects at indexes 0, 1, and 2
        """
        match len(required_data_indexes):
            case 0:
                # 0 means require all data
                return sampler  # Callable[[],SensorDataSequence]
            case 1:
                # 1 means require a specific data
                unique_index = required_data_indexes[0]
                return lambda: sampler()[unique_index]  # Callable[[], SensorData]
            case _:
                # >1 means require multiple data
                required_indexes = required_data_indexes

                def _fun() -> Tuple[SensorData, ...]:
                    data = sampler()
                    return tuple(data[i] for i in required_indexes)

                return _fun  # Callable[[],SensorDataSequence]

    @staticmethod
    def resolve_idx_sampler(sampler: IndexedSampler, required_data_indexes: Sequence[int]) -> UpdaterClosure:
        """
        Resolves the indexed sampler based on the given sampler usage.

        Args:
            sampler (IndexedSampler): The indexed sampler to resolve.
            required_data_indexes (Sequence[int]): The required data indexes.

        Returns:
            UpdaterClosure: A callable that returns a tuple of SensorData objects or a single SensorData object based on the number of required data indexes.

        Raises:
            RequirementError: If no required data indexes are specified.

        Examples:
            sampler = IndexedSampler()
            required_data_indexes = []
            resolved_sampler = resolve_idx_sampler(sampler, required_data_indexes)  # Raises RequirementError

            sampler = IndexedSampler()
            required_data_indexes = [0]
            resolved_sampler = resolve_idx_sampler(sampler, required_data_indexes)
            data = resolved_sampler()  # Returns the SensorData object at index 0

            sampler = IndexedSampler()
            required_data_indexes = [0, 1, 2]
            resolved_sampler = resolve_idx_sampler(sampler, required_data_indexes)
            data = resolved_sampler()  # Returns a tuple of the SensorData objects at indexes 0, 1, and 2
        """

        match len(required_data_indexes):
            case 0:
                raise RequirementError("Must specify at least one required data index")
            case 1:
                # 1 means require a specific data
                unique_index = required_data_indexes[0]
                return lambda: sampler(unique_index)  # Callable[[], SensorData]
            case _:
                # >1 means require multiple data
                required_indexes = required_data_indexes
                return lambda: tuple(sampler(ri) for ri in required_indexes)  # Callable[[], SensorDataSequence]

    @staticmethod
    def resolve_drc_sampler(sampler: DirectSampler, required_data_indexes: Sequence[int]) -> UpdaterClosure:
        """
        Resolves the direct sampler based on the given direct sampler and required data indexes.

        Args:
            sampler (DirectSampler): The direct sampler to resolve.
            required_data_indexes (Sequence[int]): The required data indexes.

        Returns:
            UpdaterClosure: A callable that returns a tuple of SensorData objects or a single SensorData object based on the number of required data indexes.
        """

        match len(required_data_indexes):
            case 0:
                return sampler  # Callable[[], SensorData]
            case 1:
                # 1 means require a specific data
                unique_index = required_data_indexes[0]
                return lambda: (sampler() >> unique_index) & 1
            case _:
                # >1 means require multiple data

                def _fun() -> SensorDataSequence:
                    temp_seq = sampler()
                    return tuple((temp_seq >> ri) & 1 for ri in required_data_indexes)

                return _fun



from enum import Enum
from typing import Callable

from numpy.random import choice


class NameGenerator:
    """
    Class for generating unique names.

    Attributes:
        basename: the base name
        counter: the counter
    """

    @property
    def basename(self) -> str:
        return self._basename

    @property
    def counter(self) -> int:
        return self._counter

    def __init__(self, basename: str):
        self._basename = basename
        self._counter = 0

    def __call__(self) -> str:
        self._counter += 1
        return f"{self._basename}{self._counter}"


class Multipliers(Enum):
    """
    Enum class for different multipliers.
    Notes:
        FLT: for float
        SRK: for shrink
        ENLG: for enlarge
    Attributes:
        FLT_UPPER: float around 1.0 but slightly higher
        FLT_MIDDLE: float exactly around 1.0
        FLT_LOWER: float around 1.0 but slightly lower

        SRK_L: slightly shrink
        SRK_LL: moderate shrink
        SRK_LLL: significantly shrink

        ENLG_L: slightly enlarge
        ENLG_LL: moderate enlarge
        ENLG_LLL: significantly enlarge
    """

    FLT_UPPER = (0.9, 0.925, 0.95, 1.0, 1.05, 1.08, 1.1, 1.17, 1.25)
    FLT_MIDDLE = (0.8, 0.85, 0.9, 0.95, 1.0, 1.05, 1.1, 1.17, 1.25)
    FLT_LOWER = (0.8, 0.825, 0.85, 0.875, 0.9, 0.925, 0.95, 1.0, 1.05, 1.08, 1.1)

    SRK_L = (0.65, 0.7, 0.725, 0.7375, 0.75, 0.7625, 0.775, 0.7875, 0.8)
    SRK_LL = (0.55, 0.6, 0.625, 0.65, 0.6625, 0.675, 0.6875, 0.7)
    SRK_LLL = (0.40, 0.5, 0.525, 0.55, 0.5625, 0.575, 0.5875, 0.6)

    ENLG_L = (1.25, 1.275, 1.3, 1.325, 1.35, 1.375, 1.4, 1.45, 1.5)
    ENLG_LL = (1.5, 1.525, 1.55, 1.575, 1.6, 1.625, 1.65, 1.7, 1.75)
    ENLG_LLL = (1.75, 1.775, 1.8, 1.825, 1.85, 1.875, 1.9, 1.95, 2.0, 2.1, 2.2, 2.4)


def make_multiplier_generator(mult: "Multipliers") -> Callable[[], float]:
    """
    Generates a function that returns a random float value from the given multiplier sequence.

    Parameters:
        mult (Multipliers): The multiplier enum value.

    Returns:
        Callable[[], float]: A function that returns a random float value from the multiplier sequence.
    """
    seq = mult.value

    def _make_multiplier() -> float:
        return choice(seq)

    return _make_multiplier



from typing import List, Tuple, Callable, Optional, Self, Type, TypeVar, Dict

from numpy import arange
from numpy.random import random

from ..modules.botix import MovingState, MovingTransition, __PLACE_HOLDER__

StateTransitionPack = Tuple[List[MovingState], List[MovingTransition]]

UnitType = TypeVar("UnitType", Type[MovingState], Type[MovingTransition])


class MovingChainComposer:
    """
    A class that manages the composition of moving states and transitions.

    Properties:
    - last_state (MovingState): The last state object in the chain container.
    - last_transition (MovingTransition): The last transition object in the chain container if it exists, otherwise None.
    - next_need (UnitType): The next unit type to be added to the chain container.
    """

    @property
    def last_state(self) -> MovingState:
        """
        Returns the last state object in the chain container.

        Returns:
            MovingState: The last state object in the chain container, or None if the container is empty.
        """
        container = self._chain_container[MovingState]
        return container[-1] if container else None

    @property
    def last_transition(self) -> MovingTransition:
        """
        Returns the last transition object in the chain container.

        Returns:
            MovingTransition: The last transition object in the chain container if it exists, otherwise None.
        """
        container = self._chain_container[MovingTransition]
        return container[-1] if container else None

    @property
    def next_need(self) -> UnitType:
        return self._type_container[0]

    def __init__(self) -> None:
        self._type_container: List[UnitType] = [MovingState, MovingTransition]

        self._chain_container: Dict[Type, List] = {MovingState: [], MovingTransition: []}

    def _flip(self):
        self._type_container.reverse()

    def init_container(self) -> Self:
        """
        Initializes the chain container by clearing all the lists in the `_chain_container` dictionary.
        If the `next_need` is not `MovingState`, the `_type_container` list is reversed.

        Parameters:
            self (MovingChainComposer): The instance of the class.

        Returns:
            Self: The current instance of the class.
        """
        if self.next_need != MovingState:
            self._flip()
        for cont in self._chain_container.values():
            cont.clear()
        return self

    def export_structure(self) -> StateTransitionPack:
        """
        Exports the structure of the chain container.

        Returns:
            StateTransitionPack: A tuple containing two lists, the first list contains all the MovingState objects in the chain container, and the second list contains all the MovingTransition objects in the chain container.

        Side Effects:
            - Initializes the chain container by calling the `init_container` method.
        """
        ret_pack = list(self._chain_container.get(MovingState)), list(self._chain_container.get(MovingTransition))
        self.init_container()
        return ret_pack

    def add(self, unit: MovingState | MovingTransition) -> Self:
        """
        Adds a `MovingState` or `MovingTransition` object to the chain container.

        Args:
            unit (MovingState | MovingTransition): The unit to be added to the chain container.

        Returns:
            Self: The current instance of the class.

        Raises:
            ValueError: If the type of the unit does not match the next need.
            RuntimeError: If the type of the unit is neither `MovingState` nor `MovingTransition`.
        """
        unit_type = type(unit)
        if unit_type != self.next_need:
            raise ValueError(f"Need {self.next_need}, got {unit}!")
        elif unit_type == MovingState:
            self._chain_container[MovingState].append(unit)
            if self.last_transition:
                self.last_transition.to_states[__PLACE_HOLDER__] = unit
        elif unit_type == MovingTransition:
            unit.from_states.append(self.last_state)
            self._chain_container[MovingTransition].append(unit)
        else:
            raise RuntimeError("Should never reach here!")
        self._flip()
        return self


def straight_chain(
    start_speed: int,
    end_speed: int,
    duration: float,
    power_exponent: float = 1.0,
    interval: float = 0.07,
    breaker: Optional[Callable[[], bool]] = None,
    state_on_break: Optional[MovingState] = MovingState(0),
) -> StateTransitionPack:
    """
    A function that calculates the states and transitions for a straight chain based on the input parameters.

    Args:
        start_speed (int): The initial speed of the chain.
        end_speed (int): The final speed of the chain.
        duration (float): The total duration of the chain movement.
        power_exponent (float, optional): The power exponent used in the calculation. Defaults to 1.0.
        interval (float, optional): The interval used in the calculation. Defaults to 0.07.
        breaker (Optional[Callable[[], bool]], optional): A function to determine if the transition should be broken. Defaults to None.
        state_on_break (Optional[MovingState], optional): The state to transition to if the chain is broken. Defaults to MovingState(0).

    Returns:
        StateTransitionPack: A tuple containing the list of states and transitions for the straight chain.
    """
    # Initialize the list of states with the starting state
    states: List[MovingState] = [MovingState(start_speed)]
    # Initialize an empty list for transitions
    transitions: List[MovingTransition] = []

    # Calculate the deviation in speed for uniform distribution across the duration
    deviation = end_speed - start_speed
    # Generate a sequence of speeds based on the given parameters
    speed_seq = [
        start_speed + round(deviation * x**power_exponent) for x in arange(0, 1.0, interval / (duration - interval))
    ]

    # Handle different scenarios based on whether a breaker function is provided
    match breaker:
        case None:
            # If no breaker function, create transitions without breaking condition
            for cur_speed in speed_seq:
                last_state = states[-1]
                cur_state = MovingState(cur_speed)
                cur_transition = MovingTransition(from_states=last_state, to_states=cur_state, duration=interval)
                states.append(cur_state)
                transitions.append(cur_transition)
            return states, transitions

        case break_function if callable(break_function):
            # If a breaker function is provided, create transitions that can be broken
            for cur_speed in speed_seq:
                last_state = states[-1]
                cur_state = MovingState(cur_speed)
                cur_transition = MovingTransition(
                    from_states=last_state,
                    to_states={False: cur_state, True: state_on_break},
                    duration=interval,
                    breaker=break_function,
                )
                states.append(cur_state)
                transitions.append(cur_transition)
            # Append the break state to the states list
            states.append(state_on_break)
            return states, transitions
        case _:
            # If breaker is neither None nor callable, raise an error
            raise ValueError("breaker must be callable or None")


def scanning_chain():
    """
    A function that calculates the states and transitions for a scanning chain.
    Returns:

    """
    raise NotImplementedError


def snaking_chain():
    """
    A function that calculates the states and transitions for a snaking chain.
    Returns:

    """
    raise NotImplementedError


def random_lr_turn_branch(
    start_state: MovingState,
    end_state: MovingState,
    start_state_duration: float,
    turn_speed: int,
    turn_duration: float,
    turn_left_prob: 0.5,
) -> Tuple[MovingTransition, MovingTransition]:
    """
    A function that generates random left and right turn states based on probabilities.
    It creates two transition states, a start transition, and a turn transition.

    Parameters:
        start_state (MovingState): The initial state.
        end_state (MovingState): The final state.
        start_state_duration (float): The duration of the start state.
        turn_speed (int): The speed of the turn.
        turn_duration (float): The duration of the turn.
        turn_left_prob (float): The probability of turning left.

    Returns:
        tuple: A tuple containing the start transition and the turn transition.
    """
    if not 0 < turn_left_prob < 1:
        raise ValueError("turn_speed must be between 0 and 1")

    def _die() -> bool:
        return random() > turn_left_prob

    left_turn_state = MovingState.turn("l", turn_speed)
    right_turn_state = MovingState.turn("r", turn_speed)
    start_transition = MovingTransition(
        from_states=start_state,
        to_states={False: right_turn_state, True: left_turn_state},
        duration=start_state_duration,
        breaker=_die,
    )
    turn_transition = MovingTransition(
        from_states=[left_turn_state, right_turn_state],
        to_states={__PLACE_HOLDER__: end_state},
        duration=turn_duration,
    )
    return start_transition, turn_transition
